/*
 * File:   chrono.c
 * Author: soutade
 *
 * Created on 29 octobre 2016, 14:50
 */
//#include <pic.h>


// PIC16F1828 Configuration Bit Settings

// 'C' source line config statements

#include <xc.h>

// #pragma config statements should precede project file includes.
// Use project enums instead of #define for ON and OFF.

// CONFIG1
#pragma config FOSC = INTOSC    // Oscillator Selection (INTOSC oscillator: I/O function on CLKIN pin)
#pragma config WDTE = OFF       // Watchdog Timer Enable (WDT disabled)
#pragma config PWRTE = OFF      // Power-up Timer Enable (PWRT disabled)
#pragma config MCLRE = ON       // MCLR Pin Function Select (MCLR/VPP pin function is MCLR)
#pragma config CP = OFF         // Flash Program Memory Code Protection (Program memory code protection is disabled)
#pragma config CPD = OFF        // Data Memory Code Protection (Data memory code protection is disabled)
#pragma config BOREN = OFF      // Brown-out Reset Enable (Brown-out Reset disabled)
#pragma config CLKOUTEN = OFF   // Clock Out Enable (CLKOUT function is disabled. I/O or oscillator function on the CLKOUT pin)
#pragma config IESO = ON        // Internal/External Switchover (Internal/External Switchover mode is enabled)
#pragma config FCMEN = ON       // Fail-Safe Clock Monitor Enable (Fail-Safe Clock Monitor is enabled)

// CONFIG2
#pragma config WRT = OFF        // Flash Memory Self-Write Protection (Write protection off)
#pragma config PLLEN = ON       // PLL Enable (4x PLL enabled)
#pragma config STVREN = ON      // Stack Overflow/Underflow Reset Enable (Stack Overflow or Underflow will cause a Reset)
#pragma config BORV = LO        // Brown-out Reset Voltage Selection (Brown-out Reset Voltage (Vbor), low trip point selected.)
#pragma config LVP = ON         // Low-Voltage Programming Enable (Low-voltage programming enabled)



/*
    A
   ---
F |   | B  X DP3
   ---
    G
   ---
E |   | C  X DP2
   ---
    D
 */

/*
 * Start button : RA0
 * Reset button : RA1
 */

/* 
  Warning : Logic is inverted (0 is on and 1 is off) for digits (A-F + DP)
  but is normal for digit selection (1-4)

#define LED_A    PORTAbits.RC1
#define LED_B    PORTAbits.RB6
#define LED_C    PORTCbits.RB7
#define LED_D    PORTCbits.RC6
#define LED_E    PORTBbits.RC7
#define LED_F    PORTBbits.RC0
#define LED_G    PORTBbits.RB4
#define LED_DP   PORTBbits.RC2
#define DIGIT_1  PORTCbits.RC4
#define DIGIT_2  PORTCbits.RC3
#define DIGIT_3  PORTCbits.RC5
#define DIGIT_4  PORTCbits.RB5
*/

#define PORTB_OFF 0xD0
#define PORTC_OFF 0xC7

#define LED_A_B    0
#define LED_B_B    (1 << 6)
#define LED_C_B    (1 << 7)
#define LED_D_B    0
#define LED_E_B    0
#define LED_F_B    0
#define LED_G_B    (1 << 4)
#define LED_DP_B   0
#define DIGIT_1_B  0
#define DIGIT_2_B  0
#define DIGIT_3_B  0
#define DIGIT_4_B  (1 << 5)

#define LED_A_C    (1 << 1)
#define LED_B_C    0
#define LED_C_C    0
#define LED_D_C    (1 << 6)
#define LED_E_C    (1 << 7)
#define LED_F_C    (1 << 0)
#define LED_G_C    0
#define LED_DP_C   (1 << 2)
#define DIGIT_1_C  (1 << 4)
#define DIGIT_2_C  (1 << 3)
#define DIGIT_3_C  (1 << 5)
#define DIGIT_4_C  0

/*static const char s_digit_matrix[10][2] = {
     A  B  C  D  E  F  G
    {1, 1, 1, 1, 1, 1, 0}, // 0
    {0, 1, 1, 0, 0, 0, 0}, // 1
    {1, 1, 0, 1, 1, 0, 1}, // 2
    {1, 1, 1, 1, 0, 0, 1}, // 3
    {0, 1, 1, 0, 0, 1, 1}, // 4
    {1, 0, 1, 1, 0, 1, 1}, // 5
    {1, 0, 1, 1, 1, 1, 1}, // 6
    {1, 1, 1, 0, 0, 0, 0}, // 7
    {1, 1, 1, 1, 1, 1, 1}, // 8
    {1, 1, 1, 1, 0, 1, 1}, // 9
    {1, 1, 1, 0, 1, 1, 1}, // A
    {1, 1, 1, 1, 1, 1, 1}, // B
    {1, 0, 0, 1, 1, 1, 0}, // C
    {1, 1, 1, 1, 1, 1, 0}, // D
    {1, 0, 0, 1, 1, 1, 1}, // E
    {1, 0, 0, 0, 1, 1, 1}, // F
    {0, 0, 0, 1, 1, 1, 0}, // L
};*/

#define DIGIT_L_POS 16

static const unsigned char s_digit_matrix[17][2] = {
    /* B    C  */
    {(PORTB_OFF & ~0xC0), (PORTC_OFF & ~0xC3)}, // 0
    {(PORTB_OFF & ~0xC0), (PORTC_OFF & ~0x00)}, // 1
    {(PORTB_OFF & ~0x50), (PORTC_OFF & ~0xC2)}, // 2
    {(PORTB_OFF & ~0xD0), (PORTC_OFF & ~0x42)}, // 3
    {(PORTB_OFF & ~0xD0), (PORTC_OFF & ~0x01)}, // 4
    {(PORTB_OFF & ~0x90), (PORTC_OFF & ~0x43)}, // 5
    {(PORTB_OFF & ~0x90), (PORTC_OFF & ~0xC3)}, // 6
    {(PORTB_OFF & ~0xC0), (PORTC_OFF & ~0x02)}, // 7
    {(PORTB_OFF & ~0xD0), (PORTC_OFF & ~0xC3)}, // 8
    {(PORTB_OFF & ~0xD0), (PORTC_OFF & ~0x43)}, // 9
    {(PORTB_OFF & ~0xD0), (PORTC_OFF & ~0x83)}, // A
    {(PORTB_OFF & ~0xD0), (PORTC_OFF & ~0xC3)}, // B
    {(PORTB_OFF & ~0x00), (PORTC_OFF & ~0xC3)}, // C
    {(PORTB_OFF & ~0xC0), (PORTC_OFF & ~0xC3)}, // D
    {(PORTB_OFF & ~0x10), (PORTC_OFF & ~0xC3)}, // E
    {(PORTB_OFF & ~0x10), (PORTC_OFF & ~0x83)}, // F
    {(PORTB_OFF & ~0x00), (PORTC_OFF & ~0xC1)}  // L
};

/*
 * We have a quartz @ 32768 Hz
 * No hardware prescaler is set. We set one with software to 4
 * Don't know why hardware presacaler doesn't works...
 * That means mutiplying by 4 the counter value
 * Interrupt is generated on 16 bits timer (65536 clock count)
 * Which give one counter count every (32768/4) = 8192Hz = 0,00012207s = 0.12207ms
 * For 32 light levels, we have at least
 * One digit cycle is 0.12207ms*32 = 3.9ms
 * Finally, 1 second is counted every (8192/32) = 256 digit cycles
 */
#define TARGET_FREQ        8192
#define TMRH_DEF           ((unsigned char)0xFF)
//#define TMRL_DEF           ((unsigned char)DIGIT_INTERVAL-DIGIT_ON_INTERVAL)
#define DIGIT_INTERVAL     ((unsigned char)32)
#define DIGIT_ON_INTERVAL  20    // From 4 to DIGIT_INTERVAL-1 (31)
#define DIGIT_MIN_ON_INTERVAL  ((unsigned char)4)    // From 4 to DIGIT_INTERVAL-1 (31)
#define DIGIT_MAX_ON_INTERVAL  (DIGIT_INTERVAL-1)    // From 4 to DIGIT_INTERVAL-1 (31)
#define DIGIT_CYCLE       (TARGET_FREQ/DIGIT_INTERVAL)

#define STOP_TIME             20  // 20 seconds
#define LDR_CAPTURE_TEST_TIME  4  // Capture each 4 seconds
#define LDR_CAPTURE_TIME      10  // Capture each 10 seconds

#define DEBOUNCE_TICK         99  // 99*digit cycle = ~386.1ms
#define DEBOUNCE_TICK_1S      255 // 255*digit cycle = ~994.5ms

enum { STATE_SLEEP=0, STATE_START, STATE_PAUSE, STATE_STOP, STATE_LDR_TEST };

#define FLAG_SECOND                (1 << 0)

#define DP_0 (1 << 0)
#define DP_1 (1 << 1)

// Labels
extern unsigned state_start_irq;
extern unsigned state_pause_irq;
extern unsigned state_stop_irq;
extern unsigned state_ldr_test_irq;

typedef struct {
    // State
    unsigned char state;
    unsigned char cur_time[4];
    unsigned char stop_time;
    unsigned char debounce;
    unsigned char cur_tick;
    unsigned char flags;
    unsigned char ldr_capture;
    unsigned char state_irq;
    // Display
    unsigned char cur_digit;
    unsigned char digit_on_interval;
    unsigned char digit_off_interval;
    unsigned char on;
    unsigned char dp;
} ctx_t;

static ctx_t s_ctx;

/* LDR resistor values */
static const unsigned int s_ldr_intervals[DIGIT_MAX_ON_INTERVAL-DIGIT_MIN_ON_INTERVAL] = {
    // Dark -> 4
    150, 
    166,
    180, // Low light -> 7
    200,
    240,
    300,
    360,
    420,
    480,
    540,
    600, // Smooth light -> 15
    660,
    720,
    780,
    840,
    900,
    960,
    980,
    1000, // Light -> 23
    1025,
    1050,
    1075,
    1100,
    1125,
    1150,
    1175,
    1200, // Full day light -> 31
};

void state_start();
void state_pause();
void state_stop();
void state_ldr_test();

static void memset(unsigned char* ptr, unsigned char c, unsigned size)
{
    while (size--)
        *ptr++ = c;
}

static void reset_state()
{
    T1CONbits.TMR1ON = 0;

    memset((unsigned char*)&s_ctx, 0, sizeof(s_ctx));
    
    s_ctx.stop_time = STOP_TIME;
    s_ctx.ldr_capture = LDR_CAPTURE_TIME;
    s_ctx.on = 1;
    s_ctx.state = STATE_STOP;
    s_ctx.state_irq = &state_stop_irq;
    
    PORTB = PORTB_OFF;
    PORTC = PORTC_OFF;

    PIR1bits.TMR1IF = 0;
    INTCONbits.IOCIF = 0;
}

static void update_digit_dp()
{
    s_ctx.cur_digit++;

    switch((s_ctx.cur_digit & 3))
    {
        case 0:
            PORTB = s_digit_matrix[s_ctx.cur_time[0]][0] | DIGIT_1_B;
            PORTC = s_digit_matrix[s_ctx.cur_time[0]][1] | DIGIT_1_C;
            break;
        case 1:
            PORTB = s_digit_matrix[s_ctx.cur_time[1]][0] | DIGIT_2_B;
            if (s_ctx.dp & DP_0)
                PORTC = (s_digit_matrix[s_ctx.cur_time[1]][1] | DIGIT_2_C) & ~LED_DP_C;
            else
                PORTC = s_digit_matrix[s_ctx.cur_time[1]][1] | DIGIT_2_C;
            break;
        case 2:
            PORTB = s_digit_matrix[s_ctx.cur_time[2]][0] | DIGIT_3_B;
            if (s_ctx.dp & DP_1)
                PORTC = (s_digit_matrix[s_ctx.cur_time[2]][1] | DIGIT_3_C) & ~LED_DP_C;
            else
                PORTC = s_digit_matrix[s_ctx.cur_time[2]][1] | DIGIT_3_C;
            break;
        case 3:
            PORTB = s_digit_matrix[s_ctx.cur_time[3]][0] | DIGIT_4_B;
            PORTC = s_digit_matrix[s_ctx.cur_time[3]][1] | DIGIT_4_C;
            break;
    }
}

static void update_digit()
{
    s_ctx.cur_digit++;

    switch((s_ctx.cur_digit & 3))
    {
        case 0:
            PORTB = s_digit_matrix[s_ctx.cur_time[0]][0] | DIGIT_1_B;
            PORTC = s_digit_matrix[s_ctx.cur_time[0]][1] | DIGIT_1_C;
            break;
        case 1:
            PORTB = s_digit_matrix[s_ctx.cur_time[1]][0] | DIGIT_2_B;
            PORTC = (s_digit_matrix[s_ctx.cur_time[1]][1] | DIGIT_2_C) & ~LED_DP_C;
            break;
        case 2:
            PORTB = s_digit_matrix[s_ctx.cur_time[2]][0] | DIGIT_3_B;
            PORTC = (s_digit_matrix[s_ctx.cur_time[2]][1] | DIGIT_3_C) & ~LED_DP_C;
            break;
        case 3:
            PORTB = s_digit_matrix[s_ctx.cur_time[3]][0] | DIGIT_4_B;
            PORTC = s_digit_matrix[s_ctx.cur_time[3]][1] | DIGIT_4_C;
            break;
    }
}

static void manage_ldr(void)
{
    unsigned int ldr_value = 0;
    unsigned char i;

    // Capture LDR value
    ADCON0bits.ADON = 1; // Enable ADON
    
    ADCON0bits.GO = 1;
    while (ADCON0bits.GO_nDONE);

    ADCON0bits.ADON = 0;
    
    if (s_ctx.state == STATE_LDR_TEST)
    {      
        s_ctx.dp = ADRESH; // hopefuly we can use bits 9 and 10

        ldr_value = ADRESL;

        if (ldr_value >= 100)
        {
            s_ctx.cur_time[1] = ldr_value/100;
            ldr_value -= s_ctx.cur_time[1]*100;
        }
        else
            s_ctx.cur_time[1] = 0;
        
        if (ldr_value >= 10)
        {
            s_ctx.cur_time[2] = ldr_value/10;
            ldr_value -= s_ctx.cur_time[2]*10;
        }
        else
            s_ctx.cur_time[2] = 0;

        s_ctx.cur_time[3] = ldr_value;
        
        s_ctx.ldr_capture = LDR_CAPTURE_TEST_TIME;
    }
    else
    {        
        s_ctx.ldr_capture = LDR_CAPTURE_TIME;
    }

    ldr_value = (ADRESH << 8) | ADRESL;

    // Optimization
    if (ADRESH & 2)
        i = 17;
    else if (ADRESH & 1)
        i = 8;
    else
        i = 0;
    
    for(; i<sizeof(s_ldr_intervals)/sizeof(s_ldr_intervals[0]); i++)
    {
        if (s_ldr_intervals[i] > ldr_value)
        {
            if (i) i--;
            break;
        }
    }

    // We use signed byte overflow to do 256 - X operation
    // and a software prescaler (4)
    s_ctx.digit_on_interval  = /* 256 */ - (4*(DIGIT_MIN_ON_INTERVAL+i)) ;
    s_ctx.digit_off_interval = /* 256 */ - ((4*DIGIT_INTERVAL + s_ctx.digit_on_interval)) ;
}

static void* start_button()
{
    switch (s_ctx.state)
    {
        case STATE_LDR_TEST:
        case STATE_SLEEP:
            reset_state();
            manage_ldr();
	    TMR1H = TMRH_DEF;
            TMR1L = s_ctx.digit_on_interval;
            T1CONbits.TMR1ON = 1;
        case STATE_STOP:
            return state_start;
        case STATE_START:
            return state_pause;
        default:
        case STATE_PAUSE:
            return state_start;
    }    
}

static void* reset_button()
{
    switch (s_ctx.state)
    {
        case STATE_LDR_TEST:
        case STATE_SLEEP:
            reset_state();
            manage_ldr();
	    TMR1H = TMRH_DEF;
            TMR1L = s_ctx.digit_on_interval;
            T1CONbits.TMR1ON = 1;
            return state_start;
        case STATE_START:
        case STATE_PAUSE:
            reset_state();
            manage_ldr();
            T1CONbits.TMR1ON = 1;
        default:
        case STATE_STOP:
            s_ctx.stop_time = STOP_TIME;
            return state_stop;
    }  
}

static inline void* start_ldr_test()
{
    reset_state();

    s_ctx.cur_time[0] = DIGIT_L_POS;
    T1CONbits.TMR1ON = 1;
    
    s_ctx.debounce = DEBOUNCE_TICK_1S;
    
    return state_ldr_test;
}

// 8 Hz interrupt
static void interrupt __interrupt_handler()
{
    void* new_state;

    if (PIR1bits.TMR1IF)
    {
        PIR1bits.TMR1IF = 0;
            
        TMR1H = TMRH_DEF;

        // Go to off
        if (s_ctx.on & 1)
        {
            s_ctx.on = 0;
            PORTB = PORTB_OFF;
            PORTC = PORTC_OFF;
            TMR1L += s_ctx.digit_off_interval;
            return;
        }

        // Go to on (digit cycle)
        s_ctx.on |= 1;

        TMR1L += s_ctx.digit_on_interval;

        if (s_ctx.debounce)
            s_ctx.debounce--;

        // 1 second
        // cur_tick is automatically reset due to overflow
        if (++s_ctx.cur_tick == 0)
            s_ctx.flags |= FLAG_SECOND ;
        
        // Jump to IRQ handler
        // We start @ 0x4, so high part (PCH) must not be valid
        PCL = (unsigned char) s_ctx.state_irq;
        
        /*
         * State START
         */
        asm ("_state_start_irq:");

        update_digit();

        return;

        /*
         * State PAUSE
         */
        
        asm ("_state_pause_irq:");

        if (s_ctx.flags & FLAG_SECOND)
        {
            s_ctx.dp = DP_0 | DP_1;
        }
        // Blink
        /* 154 = 0.6*digit cycle */
        /* Too much operations for multiplication */
        else if (s_ctx.dp && s_ctx.cur_tick >= 154)
        {
            s_ctx.dp = 0;
        }
        update_digit_dp();

        return;

        /*
         * State STOP
         */
        
        asm ("_state_stop_irq:");

        update_digit();

        return;

        /*
         * State LDR_TEST
         */
        
        asm ("_state_ldr_test_irq:");

        update_digit_dp();

        return;
    }

   // Start
    if (IOCAFbits.IOCAF0)
    {
        IOCAFbits.IOCAF0 = 0;

        if (s_ctx.debounce)
            return;
        s_ctx.debounce = DEBOUNCE_TICK;
      
        if (PORTAbits.RA1 == 0)
            new_state = start_ldr_test();
        else
            new_state = start_button();
    }
    // Reset
    else if (IOCAFbits.IOCAF1)
    {
        IOCAFbits.IOCAF1 = 0;

        if (s_ctx.debounce)
            return;
        s_ctx.debounce = DEBOUNCE_TICK;

        if (PORTAbits.RA0 == 0)
            new_state = start_ldr_test();
        else
            new_state = reset_button();
    }

    // Update return PC with new state
    TOSH = ((unsigned long)new_state) >> 8;
    TOSL = ((unsigned long)new_state) & 0xFF;
}

void state_start()
{
    s_ctx.state = STATE_START;
    s_ctx.state_irq = &state_start_irq;
    
    while (1)
    {
        if (s_ctx.flags & FLAG_SECOND)
        {
            if (--s_ctx.ldr_capture == 0)
                manage_ldr();

            if ((++s_ctx.cur_time[3]) > 9)
            {
                s_ctx.cur_time[3] = 0;
                if ((++s_ctx.cur_time[2]) > 5)
                {
                    s_ctx.cur_time[2] = 0;
                    if ((++s_ctx.cur_time[1]) > 9)
                    {
                        s_ctx.cur_time[1] = 0;
                        if ((++s_ctx.cur_time[0]) > 9)
                        {
                            s_ctx.cur_time[0] = 0;
                        }
                    }
                }
            }

            s_ctx.flags &= ~FLAG_SECOND;
        }
    }
}

void state_pause()
{
    s_ctx.state = STATE_PAUSE;
    s_ctx.state_irq = &state_pause_irq;

    while (1)
    {
        if (s_ctx.flags & FLAG_SECOND)
        {
            if (--s_ctx.ldr_capture == 0)
                manage_ldr();
            s_ctx.flags &= ~FLAG_SECOND;
        }
    }
}

void shutdown(void)
{
    T1CONbits.TMR1ON = 0;
    s_ctx.state = STATE_SLEEP;
    PORTB = PORTB_OFF;
    PORTC = PORTC_OFF;
    asm ("sleep");   
}

void state_stop()
{
    s_ctx.state = STATE_STOP;
    s_ctx.state_irq = &state_stop_irq;

    while (1)
    {
        if ((s_ctx.flags & FLAG_SECOND))
        {
            if (--s_ctx.stop_time == 0)
                shutdown();
            else if (--s_ctx.ldr_capture == 0)
                manage_ldr();
            s_ctx.flags &= ~FLAG_SECOND;
        }
    }
}

void state_ldr_test()
{
    s_ctx.state = STATE_LDR_TEST;
    s_ctx.state_irq = &state_ldr_test_irq;

    manage_ldr();

    while (1)
    {
        if ((s_ctx.flags & FLAG_SECOND))
        {
            if (--s_ctx.ldr_capture == 0)
                manage_ldr();
            s_ctx.flags &= ~FLAG_SECOND;
        }
    }
}

void main(void)
{
    OSCCONbits.IRCF = 12;  // 2Mhz
    OSCCONbits.SCS = 2;   // FOSC Determined by IRCF
    
    OPTION_REGbits.nWPUEN = 0; // Enable weak pull up
    
    T1CONbits.T1OSCEN = 1;  // T1 is a crystal
    T1CONbits.TMR1CS = 2;   // T1 is a crystal
    T1CONbits.T1CKPS = 0;   // No Prescaler
    T1GCONbits.TMR1GE = 0 ; // No gate function
    T1CONbits.nT1SYNC = 1;  // Asynchronous

    // Make digital
    ANSELAbits.ANSA0 = 0; // RA0 digital
    ANSELAbits.ANSA1 = 0; // RA1 digital
    ANSELB = 0;
    ANSELC = 0;

    ANSELAbits.ANSA2 = 1; // RA2 analog
    
    // Make output
    TRISB = 0;
    TRISC = 0;

    // Make input
    TRISAbits.TRISA0 = 1;
    TRISAbits.TRISA1 = 1;
    TRISAbits.TRISA2 = 1;
       
    // Interrupts on rising edge
    IOCAPbits.IOCAP0 = 1;
    IOCAPbits.IOCAP1 = 1;

    // Weak pull ups
    WPUAbits.WPUA0 = 1;
    WPUAbits.WPUA1 = 1;

    ADCON0bits.CHS = 2;   // Use RA2
    //ADCON1bits.ADCS = 6; // Max time to ADC
    ADCON1bits.ADCS = 0;   // FOSC/2
    ADCON1bits.ADPREF = 0; // Use VDD as voltage reference
    ADCON1bits.ADNREF = 0; // Use VSS as voltage reference
    ADCON1bits.ADFM = 1;   // Right justify conversion value    
    
    INTCONbits.IOCIE = 1;  // Enable interrupt on change
    PIE1bits.TMR1IE = 1;   // Enable timer 1 interrupt
    INTCONbits.PEIE = 1;   // Enable peripherals interrupt
    INTCONbits.GIE = 1;    // Enable global interrupts
    
    reset_state();
    
    shutdown();
    
    // Uncomment the next lines for debug (and comment the previous one)
    /*T1CONbits.TMR1ON = 1 ; // Activate timer 1
    manage_ldr();

    while (1)*/
        state_start();

    // Avoid optimization that removes these functions
    state_pause();
    state_stop();
    state_ldr_test();
}
